# tests/test_model.py

import numpy as np
import pytest
from numpy.testing import assert_allclose
from sklearn.datasets import load_digits, make_classification
from sklearn.metrics import accuracy_score, roc_auc_score
from sklearn.model_selection import StratifiedKFold, train_test_split
from sklearn.preprocessing import OneHotEncoder, StandardScaler
from sklearn.utils.estimator_checks import parametrize_with_checks
from ucimlrepo import fetch_ucirepo

from gfdl.model import EnsembleGFDLClassifier, GFDLClassifier

activations = ["relu", "tanh", "sigmoid", "identity", "softmax", "softmin",
               "log_sigmoid", "log_softmax"]
weights = ["zeros", "range", "uniform", "normal", "he_uniform", "lecun_uniform",
           "glorot_uniform", "he_normal", "lecun_normal", "glorot_normal"]


@pytest.mark.parametrize(
       "hidden_layer_sizes",
       [(10,), (10, 10), (5, 10, 15, 20), (100,)]
       )
@pytest.mark.parametrize("n_classes", [2, 5])
@pytest.mark.parametrize("direct_links", [0, 1])
@pytest.mark.parametrize("activation", activations)
@pytest.mark.parametrize("weight_scheme", weights)
def test_model(hidden_layer_sizes, n_classes, activation, weight_scheme, direct_links):
    N, d = 60, 10
    X, y = make_classification(n_samples=N,
                               n_features=d,
                               n_classes=n_classes,
                               n_informative=8,
                               random_state=42)

    model = GFDLClassifier(hidden_layer_sizes, activation, weight_scheme,
                           direct_links, 0)

    model.fit(X, y)

    assert len(model.W_) == len(hidden_layer_sizes)
    assert model.W_[0].T.shape == (d, hidden_layer_sizes[0])

    for layer, w, b, i in zip(
        hidden_layer_sizes[1:],
        model.W_[1:],
        model.b_[1:],
        range(len(model.W_) - 1), strict=False
        ):
        assert w.T.shape == (hidden_layer_sizes[i], layer)
        assert b.shape == (layer,)

    if direct_links:
        assert model.coeff_.shape == (
            sum(hidden_layer_sizes) + d, len(np.arange(n_classes))
            )
    else:
        assert model.coeff_.shape == (sum(hidden_layer_sizes),
                                      len(np.arange(n_classes)))

    pred = model.predict(X[:10])
    assert set(np.unique(pred)).issubset(set(np.arange(n_classes)))
    np.testing.assert_array_equal(np.unique(y), np.arange(n_classes))

    P = model.predict_proba(X[:10])
    np.testing.assert_allclose(P.sum(axis=1), 1.0, atol=1e-6)
    assert (P >= 0).all() and (P <= 1).all()
    np.testing.assert_array_equal(pred, model.classes_[np.argmax(P, axis=1)])


@pytest.mark.parametrize("weight_scheme", weights)
@pytest.mark.parametrize(
        "hidden_layer_size",
        [(10,), (2, 3, 2, 1), (5, 10, 15, 20, 15, 10), (100,)]
        )
def test_multilayer_math(weight_scheme, hidden_layer_size):

    N, d = 60, 10
    X, y = make_classification(n_samples=N,
                               n_features=d,
                               n_classes=3,
                               n_informative=8,
                               random_state=42)

    model = GFDLClassifier(
        hidden_layer_sizes=hidden_layer_size,
        activation="identity",
        weight_scheme=weight_scheme,
        direct_links=False,
        seed=0
        )

    model.fit(X, y)

    enc = OneHotEncoder(sparse_output=False, handle_unknown="ignore")
    Y = enc.fit_transform(y.reshape(-1, 1))

    # collapsing weights and biases for representation as linear operation
    weights = [w.T for w in model.W_]
    Ts, cs = [], []
    T = np.eye(X.shape[1])
    c = np.zeros((X.shape[1],))

    for w, b in zip(weights, model.b_, strict=False):
        T = T @ w
        c = c @ w + b
        Ts.append(T)
        cs.append(c)

    # design matrix with ALL layers concatenated
    expected_phi = np.hstack([X @ T_l + c_l for T_l, c_l in zip(Ts, cs, strict=False)])

    expected_beta = np.linalg.pinv(expected_phi) @ Y

    np.testing.assert_allclose(model.coeff_, expected_beta)


@pytest.mark.parametrize("hidden_layer_sizes, activation, weight_scheme, exp_auc", [
    # when direct links are absent (ELM), we expect the
    # ROC AUC to increase with multi-layer network complexity
    # up to a reasonable degree, when the width of the layers is
    # quite small
     ((2,), "relu", "uniform", 0.5598328634285958),
     ((2, 2), "relu", "uniform", 0.5666639967533855),
     # start hitting diminishing returns here:
     ((2, 2, 2, 2), "relu", "uniform", 0.5666639967533855),
     ((2, 2, 2, 2, 2, 2, 2), "relu", "uniform", 0.5666639967533855),
     # effectively no improvement here:
     ((2, 2, 2, 2, 2, 2, 2, 2, 2), "relu", "uniform", 0.5666639967533855),
     ]
 )
def test_multilayer_progression(weight_scheme,
                                hidden_layer_sizes,
                                activation,
                                exp_auc):
    X, y = make_classification(n_samples=400,
                               n_features=100,
                               n_classes=5,
                               n_informative=26,
                               random_state=42,
                               class_sep=0.5)
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2,
                                                        random_state=0)
    model = GFDLClassifier(
        hidden_layer_sizes=hidden_layer_sizes,
        activation=activation,
        weight_scheme=weight_scheme,
        direct_links=False,
        seed=0
        )
    model.fit(X_train, y_train)
    y_score = model.predict_proba(X_test)
    actual_auc = roc_auc_score(y_test, y_score, multi_class="ovo")
    assert_allclose(actual_auc, exp_auc)


@pytest.mark.parametrize(
        "Classifier, target",
        [(GFDLClassifier, 0.7161), (EnsembleGFDLClassifier, 0.7132)]
        )
def test_against_shi2021(Classifier, target):
    # test multilayer classification against
    # the results given in Shi et al. (2021) DOI 10.1016/j.patcog.2021.107978
    # dataset obtained from UCI ML repo
    abalone = fetch_ucirepo(id=1)

    X = abalone.data.features
    y = abalone.data.targets

    X = X.assign(
            Sex=lambda d: d["Sex"].map({"M": 0, "F": 1, "I": 2}).astype("int8")
            )

    X, y = np.array(X), np.array(y).reshape(-1)

    # Shi et al. only use 3 classes, but all samples are used, which implies binning
    # they do not specify the bins used, so I used my best judgement
    y = np.digitize(y, bins=[7, 11])

    # Shi et al. used half of the titanic dataset to tune
    # so I assumed they did the same for this dataset
    X_tune, X_eval, y_tune, y_eval = train_test_split(
        X, y, test_size=0.5, random_state=0)

    # Shi et al. used 4 folds on the titanic dataset
    # I inferred that they also used 4 folds for this dataset
    K = 4

    # X_tune and y_tune were used to find the hyperparameters
    # using Shi et al.'s two-stage tuning method, disregarding
    # parameter C and tuning the activation function instead
    """
    For RVFL based models, we use a two-stage tuning method to obtain
    their best hyperparameter configurations. The two-stage tuning can be
    performed by the following steps: 1) Fix the number layers to 2, and
    then select the optimal number of neurons (N*) and regularization
    parameter (C*) using a coarse range for N and C. 2) Tune the number
    of layers and fine tune the N, C parameters by considering only a fine
    range in the neighborhood of N* and C*.

    Shi et al. (2021) https://doi.org/10.1016/j.patcog.2021.107978
    """

    # values determined using method outlined above
    hidden_layer_sizes = [512, 512]
    reg = 16

    model = Classifier(
        hidden_layer_sizes=hidden_layer_sizes,
        activation="relu",
        weight_scheme="uniform",
        reg_alpha=reg,
        seed=0
        )

    scl = StandardScaler()

    # The actual splits used in the paper were not specified
    skf = StratifiedKFold(n_splits=K, shuffle=True, random_state=42)

    acc = 0
    for train_index, test_index in skf.split(X_eval, y_eval):
        X_train = X_eval[train_index]
        y_train = y_eval[train_index]
        X_test = X_eval[test_index]
        y_test = y_eval[test_index]

        model.fit(scl.fit_transform(X_train), y_train)

        y_hat = model.predict(scl.transform(X_test))

        acc += accuracy_score(y_test, y_hat)

    acc /= K

    # not an exact match because they don't specify their activation
    # nor do they mention the best hyperparameter configuration
    # and they're using ridge

    # tightest bound for both rel and abs
    # values in paper:
    # dRVFL accuracy: 66.33%
    # edRVFL accuracy: 65.81%
    assert acc == pytest.approx(target, rel=1e-4, abs=0)


def test_soft_and_hard():
    N, d = 60, 10
    X, y = make_classification(n_samples=N,
                               n_features=d,
                               n_classes=3,
                               n_informative=8,
                               random_state=0)

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2,
                                                        random_state=0)

    model = EnsembleGFDLClassifier(
        hidden_layer_sizes=(5, 5, 5),
        activation="tanh",
        weight_scheme="uniform",
        seed=0,
        reg_alpha=0.1
    )
    model.fit(X_train, y_train)

    y_soft = model.predict(X_test)

    P = model.predict_proba(X_test)
    y_from_mean = model.classes_[np.argmax(P, axis=1)]
    np.testing.assert_equal(y_soft, y_from_mean)

    model.voting = "hard"
    y_hard = model.predict(X_test)

    np.testing.assert_equal(y_soft, y_hard)


def test_hard_vote_proba_error():
    X, y = make_classification(n_samples=60,
                               n_features=10,
                               n_classes=3,
                               n_informative=8,
                               random_state=0)
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2,
                                                        random_state=0)
    model = EnsembleGFDLClassifier(
        hidden_layer_sizes=(5, 5, 5),
        activation="tanh",
        weight_scheme="uniform",
        seed=0,
        reg_alpha=0.1,
        voting="hard",
    )
    model.fit(X_train, y_train)
    with pytest.raises(AttributeError, match="predict_proba"):
        model.predict_proba(X_test)


@pytest.mark.parametrize("alpha", [None, 0.1])
def test_soft_and_hard_can_differ(alpha):
    N, d = 60, 10
    X, y = make_classification(n_samples=N,
                               n_features=d,
                               n_classes=3,
                               n_informative=8,
                               random_state=0)

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2,
                                                        random_state=0)

    # adding more layers (heads) increases the chance of disagreement
    # between the two voting methods
    model = EnsembleGFDLClassifier(
        hidden_layer_sizes=(3, 3, 3, 3),
        activation="tanh",
        weight_scheme="uniform",
        seed=0,
        reg_alpha=alpha
    )
    model.fit(X_train, y_train)
    y_soft = model.predict(X_test)
    model.voting = "hard"
    y_hard = model.predict(X_test)
    difference = [
        True, True, True, True, True, True, True, True, True, True, False, True
        ]

    np.testing.assert_array_equal(y_soft == y_hard, difference)


@pytest.mark.parametrize("Classifier", [GFDLClassifier, EnsembleGFDLClassifier])
def test_invalid_activation_weight(Classifier):
    X = np.zeros((30, 4))
    y = np.zeros((30,))
    invalid_act = Classifier(hidden_layer_sizes=100,
                             activation="bogus_activation",
                             weight_scheme="uniform")
    invalid_weight = Classifier(hidden_layer_sizes=100,
                                activation="identity",
                                weight_scheme="bogus_weight")
    # the sklearn estimator API bans input validation in __init__,
    # so we need to call fit() for error handling to kick in:
    # https://scikit-learn.org/stable/developers/develop.html#developing-scikit-learn-estimators
    with pytest.raises(ValueError, match="is not supported"):
        invalid_act.fit(X, y)
    with pytest.raises(ValueError, match="is not supported"):
        invalid_weight.fit(X, y)


@pytest.mark.parametrize("Classifier", [GFDLClassifier, EnsembleGFDLClassifier])
def test_invalid_alpha(Classifier):
    # the sklearn estimator API bans input validation in __init__,
    # so we need to call fit() for error handling to kick in:
    # https://scikit-learn.org/stable/developers/develop.html#developing-scikit-learn-estimators
    X = np.zeros((30, 4))
    y = np.zeros((30,))
    bad_est = Classifier(hidden_layer_sizes=100,
                             activation="identity",
                             weight_scheme="uniform",
                             reg_alpha=-10)
    with pytest.raises(ValueError, match=r"Negative reg\_alpha"):
        bad_est.fit(X, y)


@pytest.mark.parametrize("""hidden_layer_sizes,
                            n_classes,
                            activation,
                            weight_scheme,
                            alpha,
                            exp_proba_shape,
                            exp_proba_median,
                            exp_proba_min""", [

                    # expected values are from graforvfl library
                    ([10,], 2, "relu", "uniform", None, (20, 2), 0.5, 0.0444571694),
                    ([100,], 2, "tanh", "normal", None, (20, 2), 0.5, 0.02538905725),
                    ([10,], 5, "softmax", "lecun_uniform", None, (20, 5),
                     0.186506112, 0.08469873),
                    ([10,], 2, "relu", "uniform", 0.5, (20, 2), 0.49999999999999994,
                    0.04676933232591643),
                    ([10,], 2, "relu", "normal", 0.5, (20, 2), 0.5,
                    0.13832596541020634),
                    ([10,], 2, "relu", "he_uniform", 0.5, (20, 2), 0.5,
                    0.09354846081377409),
                    ([10,], 2, "relu", "lecun_uniform", 0.5, (20, 2), 0.5,
                    0.09387932375067173),
                    ([10,], 2, "relu", "glorot_uniform", 0.5, (20, 2),
                    0.49999999999999994, 0.09474642560519067),
                    ([10,], 2, "relu", "he_normal", 0.5, (20, 2), 0.5,
                    0.13756805074436051),
                    ([10,], 2, "relu", "lecun_normal", 0.5, (20, 2), 0.5,
                    0.1366715193146648),
                    ([10,], 2, "relu", "glorot_normal", 0.5, (20, 2), 0.5,
                    0.147434110768701),
                    ([100,], 5, "relu", "normal", 1, (20, 5), 0.15697278777061396,
                    0.014480242978774488),
                    ([100,], 5, "tanh", "normal", 1, (20, 5), 0.18173657135483476,
                    0.04755723146401269),
                    ([100,], 5, "sigmoid", "normal", 1, (20, 5), 0.1831653950464296,
                    0.05378741996708733),
                    ([100,], 5, "softmax", "normal", 1, (20, 5), 0.19357646668265396,
                     0.10898717209741866),
                    ([100,], 5, "softmin", "normal", 1, (20, 5), 0.18746771358297387,
                     0.09186562406164228),
                    ([100,], 5, "log_sigmoid", "normal", 1, (20, 5),
                    0.16722029352468032, 0.012690348255702557),
                    ([100,], 5, "log_softmax", "normal", 1, (20, 5),
                    0.1853363666712296, 0.10846041127337658),
])
def test_classification_against_grafo(hidden_layer_sizes, n_classes, activation,
                                      weight_scheme, alpha, exp_proba_shape,
                                      exp_proba_median, exp_proba_min):
    # test binary and multi-class classification against expected values
    # from the open source graforvfl library on some synthetic
    # datasets
    X, y = make_classification(n_classes=n_classes,
                               n_informative=8, random_state=0)
    X_train, X_test, y_train, _ = train_test_split(X, y, test_size=0.2,
                                                        random_state=0)
    model = GFDLClassifier(hidden_layer_sizes=hidden_layer_sizes,
                 activation=activation,
                 weight_scheme=weight_scheme,
                 direct_links=1,
                 seed=0,
                 reg_alpha=alpha)
    model.fit(X_train, y_train)

    actual_proba = model.predict_proba(X_test)
    actual_proba_shape = actual_proba.shape
    actual_proba_median = np.median(actual_proba)
    actual_proba_min = np.min(actual_proba)

    np.testing.assert_allclose(actual_proba_shape, exp_proba_shape)
    np.testing.assert_allclose(actual_proba_median, exp_proba_median)
    np.testing.assert_allclose(actual_proba_min, exp_proba_min)


@parametrize_with_checks([GFDLClassifier(), EnsembleGFDLClassifier()])
def test_sklearn_api_conformance(estimator, check):
    check(estimator)


@pytest.mark.parametrize("reg_alpha, rtol, expected_acc, expected_roc", [
    (0.1, 1e-15, 0.9083333333333333, 0.9893414717354735),
    (None, 1e-15, 0.2222222222222222, 0.5518850599798965),
    (None, 1e-3, 0.8972222222222223, 0.9802912857599967),
])
def test_rtol_classifier(reg_alpha, rtol, expected_acc, expected_roc):
    # For Moore-Penrose, a large singular value cutoff (rtol)
    # may be required to achieve reasonable results. This test
    # showcases that a default low cut off leads to almost random classification
    # output for the Digits datasets which is alleviated by increasing the cut off.
    # This cut off has no effect on ridge solver.
    data = load_digits()
    X, y = data.data, data.target
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2,
                                                        random_state=0)

    scaler = StandardScaler().fit(X_train)
    X_train_s = scaler.transform(X_train)
    X_test_s = scaler.transform(X_test)

    model = GFDLClassifier(hidden_layer_sizes=[800] * 10,
            activation="softmax",
            weight_scheme="normal",
            seed=0,
            reg_alpha=reg_alpha,
            rtol=rtol)
    model.fit(X_train_s, y_train)

    y_hat_cur = model.predict(X_test_s)
    y_hat_cur_proba = model.predict_proba(X_test_s)

    acc_cur = accuracy_score(y_test, y_hat_cur)
    roc_cur = roc_auc_score(y_test, y_hat_cur_proba, multi_class="ovo")

    np.testing.assert_allclose(acc_cur, expected_acc)
    np.testing.assert_allclose(roc_cur, expected_roc)


@pytest.mark.parametrize("reg_alpha, rtol, expected_acc, expected_roc", [
    (5.0, 1e-15, 0.7222222222222222, 0.9525486362311113),
    (None, 1e-15, 0.10833333333333334, 0.5062846049300238),
    (None, 1e-3, 0.9555555555555556, 0.9920190654177233),
])
def test_rtol_ensemble(reg_alpha, rtol, expected_acc, expected_roc):
    # For Moore-Penrose, a large singular value cutoff (rtol)
    # may be required to achieve reasonable results. This test
    # showcases that a default low cut off leads to almost random classification
    # output for the Digits datasets which is alleviated by increasing the cut off.
    # This cut off has no effect on ridge solver.
    data = load_digits()
    X, y = data.data, data.target
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2,
                                                        random_state=0)

    scaler = StandardScaler().fit(X_train)
    X_train_s = scaler.transform(X_train)
    X_test_s = scaler.transform(X_test)

    model = EnsembleGFDLClassifier(hidden_layer_sizes=[2000] * 2,
            activation="relu",
            weight_scheme="uniform",
            seed=0,
            reg_alpha=reg_alpha,
            rtol=rtol)
    model.fit(X_train_s, y_train)

    y_hat_cur = model.predict(X_test_s)
    y_hat_cur_proba = model.predict_proba(X_test_s)

    acc_cur = accuracy_score(y_test, y_hat_cur)
    roc_cur = roc_auc_score(y_test, y_hat_cur_proba, multi_class="ovo")

    np.testing.assert_allclose(acc_cur, expected_acc)
    np.testing.assert_allclose(roc_cur, expected_roc, atol=1e-05)



"""
Estimators for gradient free deep learning.
"""

import numpy as np
from scipy.special import logsumexp
from scipy.stats import mode
from sklearn.base import (
    BaseEstimator,
    ClassifierMixin,
    MultiOutputMixin,
    RegressorMixin,
)
from sklearn.linear_model import Ridge
from sklearn.preprocessing import OneHotEncoder
from sklearn.utils.metaestimators import available_if
from sklearn.utils.multiclass import unique_labels
from sklearn.utils.validation import check_is_fitted, validate_data

from gfdl.activations import resolve_activation
from gfdl.weights import resolve_weight


class GFDL(BaseEstimator):
    """Base class for GFDL for classification and regression."""
    def __init__(
        self,
        hidden_layer_sizes: np.typing.ArrayLike = (100,),
        activation: str = "identity",
        weight_scheme: str = "uniform",
        direct_links: bool = True,
        seed: int = None,
        reg_alpha: float = None,
        rtol: float | None = None,
    ):
        self.hidden_layer_sizes = hidden_layer_sizes
        self.activation = activation
        self.direct_links = direct_links
        self.seed = seed
        self.weight_scheme = weight_scheme
        self.reg_alpha = reg_alpha
        self.rtol = rtol

    def fit(self, X, Y):
        # Assumption : X, Y have been pre-processed.
        # X shape: (n_samples, n_features)
        # Y shape: (n_samples, n_classes-1)
        if self.reg_alpha is not None and self.reg_alpha < 0.0:
            raise ValueError("Negative reg_alpha. Expected range : None or [0.0, inf).")
        fn = resolve_activation(self.activation)[1]
        self._activation_fn = fn
        self._N = X.shape[1]
        hidden_layer_sizes = np.asarray(self.hidden_layer_sizes)
        self._weight_mode = resolve_weight(self.weight_scheme)

        # weights shape: (n_layers,)
        # biases shape: (n_layers,)
        self.W_ = []
        self.b_ = []
        rng = self.get_generator(self.seed)

        self.W_.append(
            self._weight_mode(
                self._N, hidden_layer_sizes[0], rng=self.get_generator(self.seed)
                )
            )
        self.b_.append(
            self._weight_mode(1, hidden_layer_sizes[0], rng=rng)
            .reshape(-1)
            )
        for i, layer in enumerate(hidden_layer_sizes[1:]):
            # (n_hidden, n_features)
            self.W_.append(
                self._weight_mode(hidden_layer_sizes[i], layer, rng=rng,)
                )
            # (n_hidden,)
            self.b_.append(
                self._weight_mode(1, layer, rng=rng,).reshape(-1)
                )

        # hypothesis space shape: (n_layers,)
        Hs = []
        H_prev = X
        for w, b in zip(self.W_, self.b_, strict=False):
            Z = H_prev @ w.T + b  # (n_samples, n_hidden)
            H_prev = self._activation_fn(Z)
            Hs.append(H_prev)

        # design matrix shape: (n_samples, sum_hidden+n_features)
        # or (n_samples, sum_hidden)
        if self.direct_links:
            Hs.append(X)
        D = np.hstack(Hs)

        # beta shape: (sum_hidden+n_features, n_classes-1)
        # or (sum_hidden, n_classes-1)

        # If reg_alpha is None, use direct solve using
        # MoorePenrose Pseudo-Inverse, otherwise use ridge regularized form.
        if self.reg_alpha is None:
            self.coeff_ = np.linalg.pinv(D, rtol=self.rtol) @ Y
        else:
            ridge = Ridge(alpha=self.reg_alpha, fit_intercept=False)
            ridge.fit(D, Y)
            self.coeff_ = ridge.coef_.T
        return self

    def predict(self, X):
        check_is_fitted(self)
        Hs = []
        H_prev = X
        for W, b in zip(self.W_, self.b_, strict=False):
            Z = H_prev @ W.T + b  # (n, m)
            H_prev = self._activation_fn(Z)
            Hs.append(H_prev)

        if self.direct_links:
            Hs.append(X)
        D = np.hstack(Hs)
        out = D @ self.coeff_

        return out

    def get_generator(self, seed):
        return np.random.default_rng(seed)


class GFDLClassifier(ClassifierMixin, GFDL):
    """
    Random vector functional link network classifier.

    This model fits a feedforward neural network with fixed random hidden-layer
    parameters and solves for the output weights using linear least squares or
    ridge regression. When direct links are disabled, the model architecture corresponds
    to an Extreme Learning Machine (ELM) architecture.

    Parameters
    ----------
    hidden_layer_sizes : array-like of shape (n_layers,), default=(100,)
        The ith element represents the number of neurons in the ith
        hidden layer.

    activation : str, default='identity'
        Activation function for the hidden layers.

        - 'identity', no-op activation, useful to implement linear bottleneck,
          returns f(x) = x

        - 'tanh': :func:`tanh <gfdl.activations.tanh>`.

        - 'relu': :func:`relu <gfdl.activations.relu>`.

        - 'sigmoid': :func:`sigmoid <gfdl.activations.sigmoid>`.

        - 'softmax': :func:`softmax <gfdl.activations.softmax>`.

        - 'softmin': :func:`softmin <gfdl.activations.softmin>`.

        - 'log_sigmoid': :func:`log_sigmoid <gfdl.activations.log_sigmoid>`.

        - 'log_softmax': :func:`log_softmax <gfdl.activations.log_softmax>`.

    weight_scheme : str, default='uniform'
        Distribution used to initialize the random hidden-layer weights.

        The initialization functions generate weight matrices of shape
        (n_hidden_units, n_features), where values are drawn
        according to the selected scheme.

        - 'zeros': set weights to zeros (:func:`zeros <gfdl.weights.zeros>`).

        - 'range': set weights to normalized np.arange
          (:func:`range <gfdl.weights.range>`).

        - 'uniform': uniform distribution (:func:`uniform <gfdl.weights.uniform>`).

        - 'he_uniform': He uniform distribution
          (:func:`he_uniform <gfdl.weights.he_uniform>`).

        - 'lecun_uniform': Lecun uniform distribution
          (:func:`lecun_uniform <gfdl.weights.lecun_uniform>`).

        - 'glorot_uniform': Glorot uniform distribution
          (:func:`glorot_uniform <gfdl.weights.glorot_uniform>`).

        - 'normal': Normal distribution (:func:`normal <gfdl.weights.normal>`).

        - 'he_normal': He normal distribution
          (:func:`he_normal <gfdl.weights.he_normal>`).

        - 'lecun_normal': Lecun normal distribution
          (:func:`lecun_normal <gfdl.weights.lecun_normal>`).

        - 'glorot_normal': Glorot normal distribution
          (:func:`glorot_normal <gfdl.weights.glorot_normal>`).

    direct_links : bool, default=True
        Whether to connect input layer to output nodes.
        When set to False, only the hidden-layer activations are used, corresponding
        to the Extreme Learning Machine (ELM) architecture.

    seed : int, RandomState instance, default=None
        Determines random number generation for weights and bias
        initialization.
        Pass an int for reproducible results across multiple function calls.
        See :term:`Glossary <random_state>`.

    reg_alpha : float, default=None
        Amount of ridge shrinkage to apply in order to improve
        conditioning during Ridge regression. When set to zero or `None`,
        model uses direct solve using Moore-Penrose Pseudo-Inverse.

    rtol : float, default=None
      Cutoff for small singular values for the Moore-Penrose
      pseudo-inverse. Only applies when ``reg_alpha=None``.
      When ``rtol=None``, the array API standard default for
      ``pinv`` is used.

    Attributes
    ----------
    n_features_in_ : int
        Number of features seen during :term:`fit`.

    classes_ : ndarray or list of ndarray of shape (n_classes,)
        Class labels for each output.

    W_ : list of ndarray of shape (n_layers,)
        Weight matrices of the hidden layers. The ith element in the list represents the
        weight matrix corresponding to layer i.

    b_ : list of ndarray of shape (n_layers,)
        Bias vectors of the hidden layers. The ith element in the list represents the
        bias term corresponding to layer i.

    coeff_ : ndarray of shape (n_features_out, n_outputs)
        Output weight matrix learned by fit method.

    See Also
    --------
    GFDLRegressor : Regressor variant for the RVFL architecture.

    Examples
    --------
    >>> from gfdl.model import GFDLClassifier
    >>> from sklearn.datasets import make_classification
    >>> from sklearn.model_selection import train_test_split
    >>> X, y = make_classification(n_samples=100, random_state=1)
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y, stratify=y,
    ...                                                     random_state=1)
    >>> clf = GFDLClassifier(seed=1).fit(X_train, y_train)
    >>> clf.predict_proba(X_test[:1])
    array([[0.46123716, 0.53876284]])
    >>> clf.predict(X_test[:5, :])
    array([1, 0, 1, 0, 1])
    """
    def __init__(
        self,
        hidden_layer_sizes: np.typing.ArrayLike = (100,),
        activation: str = "identity",
        weight_scheme: str = "uniform",
        direct_links: bool = True,
        seed: int = None,
        reg_alpha: float = None,
        rtol: float = None
    ):
        super().__init__(hidden_layer_sizes=hidden_layer_sizes,
                       activation=activation,
                       weight_scheme=weight_scheme,
                       direct_links=direct_links,
                       seed=seed,
                       reg_alpha=reg_alpha,
                       rtol=rtol)

    def fit(self, X, y):
        """
        Build a gradient-free neural network from the training set (X, y).

        Parameters
        ----------

        X : array-like of shape (n_samples, n_features)
          The training input samples.
        y : array-like of shape (n_samples,) or (n_samples, n_outputs)
          The target values (class labels).

        Returns
        -------
        object
          Fitted estimator.
        """
        # shape: (n_samples, n_features)
        X, Y = validate_data(self, X, y)
        self.classes_ = unique_labels(Y)

        # onehot y
        # (this is necessary for everything beyond binary classification)
        self.enc_ = OneHotEncoder(handle_unknown="ignore", sparse_output=False)
        # shape: (n_samples, n_classes-1)
        Y = self.enc_.fit_transform(Y.reshape(-1, 1))

        # call base fit method
        super().fit(X, Y)
        return self

    def predict(self, X):
        """
        Predict class for X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
          The input samples.

        Returns
        -------
        ndarray
          The predicted classes, with shape (n_samples,) or (n_samples, n_outputs).
        """
        check_is_fitted(self)
        X = validate_data(self, X, reset=False)
        out = self.predict_proba(X)
        y_hat = self.classes_[np.argmax(out, axis=1)]
        return y_hat

    def predict_proba(self, X):
        """
        Predict class probabilities for X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
          The input samples.

        Returns
        -------
        ndarray
          The class probabilities of the input samples. The order of the classes
          corresponds to that in the attribute ``classes_``. The ndarray should
          have shape (n_samples, n_classes).
        """
        check_is_fitted(self)
        X = validate_data(self, X, reset=False)
        out = super().predict(X)
        out = np.exp(out - logsumexp(out, axis=1, keepdims=True))
        return out


class EnsembleGFDL(GFDL):
    """Base class for ensemble GFDL model for classification and regression."""
    def __init__(
        self,
        hidden_layer_sizes: np.typing.ArrayLike = (100,),
        activation: str = "identity",
        weight_scheme: str = "uniform",
        seed: int = None,
        reg_alpha: float = None,
        rtol: float | None = None,
    ):
        super().__init__(hidden_layer_sizes=hidden_layer_sizes,
                         activation=activation,
                         weight_scheme=weight_scheme,
                         direct_links=True,
                         seed=seed,
                         reg_alpha=reg_alpha,
                         rtol=rtol)

    def fit(self, X, Y):

        if self.reg_alpha is not None and self.reg_alpha < 0.0:
            raise ValueError("Negative reg_alpha. Expected range : None or [0.0, inf).")

        fn = resolve_activation(self.activation)[1]
        self._activation_fn = fn
        self._N = X.shape[1]
        hidden_layer_sizes = np.asarray(self.hidden_layer_sizes)
        self._weight_mode = resolve_weight(self.weight_scheme)

        self.W_ = []
        self.b_ = []
        rng = self.get_generator(self.seed)

        self.W_.append(
            self._weight_mode(
                self._N, hidden_layer_sizes[0], rng=self.get_generator(self.seed)
                )
            )
        self.b_.append(
            self._weight_mode(1, hidden_layer_sizes[0], rng=rng)
            .reshape(-1)
            )

        for i, layer in enumerate(hidden_layer_sizes[1:]):
            # (n_hidden, n_features)
            self.W_.append(
                self._weight_mode(hidden_layer_sizes[i] + self._N, layer, rng=rng)
                )
            # (n_hidden,)
            self.b_.append(
                self._weight_mode(1, layer, rng=rng,).reshape(-1)
                )

        self.coeffs_ = []
        D = X

        for W, b in zip(self.W_, self.b_, strict=False):
            Z = D @ W.T + b  # (n_samples, n_hidden_layer_i)
            H = self._activation_fn(Z)
            # design matrix shape: (n_samples, n_hidden_layer_i+n_features)
            # or (n_samples, n_hidden_final)
            D = np.hstack((H, X))

            # beta shape: (n_hidden_final+n_features, n_classes-1)
            # or (n_hidden_final, n_classes-1)

            # If reg_alpha is None, use direct solve using
            # MoorePenrose Pseudo-Inverse, otherwise use ridge regularized form.
            if self.reg_alpha is None:
                coeff = np.linalg.pinv(D, rtol=self.rtol) @ Y
            else:
                ridge = Ridge(alpha=self.reg_alpha, fit_intercept=False)
                ridge.fit(D, Y)
                coeff = ridge.coef_.T
            self.coeffs_.append(coeff)

        return self

    def _forward(self, X):
        check_is_fitted(self)
        outs = []

        D = X
        for W, b, coeff in zip(self.W_, self.b_, self.coeffs_, strict=True):
            Z = D @ W.T + b
            H = self._activation_fn(Z)

            D = np.hstack((H, X))

            out = D @ coeff
            outs.append(out)

        return outs


class EnsembleGFDLClassifier(ClassifierMixin, EnsembleGFDL):
    """
    Ensemble random vector functional link network classifier.

    Parameters
    ----------

    hidden_layer_sizes : array-like of shape (n_layers,)
      The ith element represents the number of neurons in the ith
      hidden layer.

    activation : str, default='identity'
        Activation function for the hidden layers.

        - 'identity', no-op activation, useful to implement linear bottleneck,
          returns f(x) = x

        - 'tanh': :func:`tanh <gfdl.activations.tanh>`.

        - 'relu': :func:`relu <gfdl.activations.relu>`.

        - 'sigmoid': :func:`sigmoid <gfdl.activations.sigmoid>`.

        - 'softmax': :func:`softmax <gfdl.activations.softmax>`.

        - 'softmin': :func:`softmin <gfdl.activations.softmin>`.

        - 'log_sigmoid': :func:`log_sigmoid <gfdl.activations.log_sigmoid>`.

        - 'log_softmax': :func:`log_softmax <gfdl.activations.log_softmax>`.

    weight_scheme : str, default='uniform'
        Distribution used to initialize the random hidden-layer weights.

        The initialization functions generate weight matrices of shape
        (n_hidden_units, n_features), where values are drawn
        according to the selected scheme.

        - 'zeros': set weights to zeros (:func:`zeros <gfdl.weights.zeros>`).

        - 'range': set weights to normalized np.arange
          (:func:`range <gfdl.weights.range>`).

        - 'uniform': uniform distribution (:func:`uniform <gfdl.weights.uniform>`).

        - 'he_uniform': He uniform distribution
          (:func:`he_uniform <gfdl.weights.he_uniform>`).

        - 'lecun_uniform': Lecun uniform distribution
          (:func:`lecun_uniform <gfdl.weights.lecun_uniform>`).

        - 'glorot_uniform': Glorot uniform distribution
          (:func:`glorot_uniform <gfdl.weights.glorot_uniform>`).

        - 'normal': Normal distribution (:func:`normal <gfdl.weights.normal>`).

        - 'he_normal': He normal distribution
          (:func:`he_normal <gfdl.weights.he_normal>`).

        - 'lecun_normal': Lecun normal distribution
          (:func:`lecun_normal <gfdl.weights.lecun_normal>`).

        - 'glorot_normal': Glorot normal distribution
          (:func:`glorot_normal <gfdl.weights.glorot_normal>`).

    seed : int, default=`None`
      Random seed used to initialize the network.

    reg_alpha : float, default=`None`
      When `None`, use Moore-Penrose inversion to solve for the output
      weights of the network. Otherwise, it specifies the constant that
      multiplies the L2 term of `sklearn` `Ridge`, controlling the
      regularization strength. `reg_alpha` must be a non-negative float.

    rtol : float, default=None
      Cutoff for small singular values for the Moore-Penrose
      pseudo-inverse. Only applies when ``reg_alpha=None``.
      When ``rtol=None``, the array API standard default for
      ``pinv`` is used.

    voting : str, default=`"soft"`
      Whether to use soft or hard voting in the ensemble.

    Notes
    -----
    The implementation is based on the one described by Shi et al. in [1]_.

    References
    ----------
    .. [1] Shi, Katuwal, Suganthan, Tanveer, "Random vector functional
       link neural network based ensemble deep learning." Pattern Recognition,
       vol. 117, pp. 107978, 2021, https://doi.org/10.1016/j.patcog.2021.107978.

    Examples
    --------
    >>> from sklearn.datasets import make_classification
    >>> from gfdl.model import EnsembleGFDLClassifier
    >>> X, y = make_classification(n_samples=1000, n_features=4,
    ...                            n_informative=2, n_redundant=0,
    ...                            random_state=0, shuffle=False)
    >>> clf = EnsembleGFDLClassifier(seed=0)
    >>> clf.fit(X, y)
    >>> print(clf.predict([[0, 0, 0, 0]]))
    [1]
    """
    def __init__(
        self,
        hidden_layer_sizes: np.typing.ArrayLike = (100,),
        activation: str = "identity",
        weight_scheme: str = "uniform",
        seed: int = None,
        reg_alpha: float = None,
        rtol: float = None,
        voting: str = "soft",    # "soft" or "hard"
    ):
        super().__init__(hidden_layer_sizes=hidden_layer_sizes,
                         activation=activation,
                         weight_scheme=weight_scheme,
                         seed=seed,
                         reg_alpha=reg_alpha,
                         rtol=rtol
                         )
        self.voting = voting

    def fit(self, X, y):
        """
        Train the ensemble of connected RVFL networks on the training set (X, y).

        Parameters
        ----------

        X : array-like of shape (n_samples, n_features)
          The training input samples.
        y : array-like of shape (n_samples,) or (n_samples, n_outputs)
          The target values.

        Returns
        -------
        object
          The fitted estimator.
        """
        # shape: (n_samples, n_features)
        X, Y = validate_data(self, X, y)
        self.classes_ = unique_labels(Y)

        # onehot y
        # (this is necessary for everything beyond binary classification)
        self.enc_ = OneHotEncoder(handle_unknown="ignore", sparse_output=False)
        # shape: (n_samples, n_classes-1)
        Y = self.enc_.fit_transform(Y.reshape(-1, 1))

        # call base fit method
        super().fit(X, Y)
        return self

    def _check_voting(self):
        if self.voting == "hard":
            raise AttributeError(
                f"predict_proba is not available when voting={self.voting!r}"
            )
        return True

    @available_if(_check_voting)
    def predict_proba(self, X):
        """
        Predict class probabilities for X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
          The input samples.

        Returns
        -------
        ndarray
          The class probabilities of the input samples. The order of the classes
          corresponds to that in the attribute ``classes_``. The ndarray should
          have shape (n_samples, n_classes).
        """
        check_is_fitted(self)
        X = validate_data(self, X, reset=False)

        outs = self._forward(X)
        probs = []

        for out in outs:
            p = np.exp(out - logsumexp(out, axis=1, keepdims=True))
            probs.append(p)

        return np.mean(probs, axis=0)

    def predict(self, X):
        """
        Predict class for X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
          The input samples.

        Returns
        -------
        ndarray
          The predicted classes, with shape (n_samples,) or (n_samples, n_outputs).
        """
        check_is_fitted(self)
        X = validate_data(self, X, reset=False)

        if self.voting == "soft":
            P = self.predict_proba(X)
            return self.classes_[np.argmax(P, axis=1)]

        outs = self._forward(X)
        votes = []

        for out in outs:
            p = np.exp(out - logsumexp(out, axis=1, keepdims=True))
            votes.append(self.classes_[np.argmax(p, axis=1)])

        votes = np.stack(votes, axis=1)
        m = mode(votes, axis=1, keepdims=False)
        return m.mode


class GFDLRegressor(RegressorMixin, MultiOutputMixin, GFDL):
    """
    Random vector functional link network regressor.

    This model fits a feedforward neural network with fixed random hidden-layer
    parameters and solves for the output weights using linear least squares or
    ridge regression. When direct links are disabled, the model architecture corresponds
    to an Extreme Learning Machine (ELM) architecture.

    Parameters
    ----------
    hidden_layer_sizes : array-like of shape (n_layers,), default=(100,)
        The ith element represents the number of neurons in the ith
        hidden layer.

    activation : str, default='identity'
        Activation function for the hidden layers.

        - 'identity', no-op activation, useful to implement linear bottleneck,
          returns f(x) = x

        - 'tanh': :func:`tanh <gfdl.activations.tanh>`.

        - 'relu': :func:`relu <gfdl.activations.relu>`.

        - 'sigmoid': :func:`sigmoid <gfdl.activations.sigmoid>`.

        - 'softmax': :func:`softmax <gfdl.activations.softmax>`.

        - 'softmin': :func:`softmin <gfdl.activations.softmin>`.

        - 'log_sigmoid': :func:`log_sigmoid <gfdl.activations.log_sigmoid>`.

        - 'log_softmax': :func:`log_softmax <gfdl.activations.log_softmax>`.

    weight_scheme : str, default='uniform'
        Distribution used to initialize the random hidden-layer weights.

        The initialization functions generate weight matrices of shape
        (n_hidden_units, n_features), where values are drawn
        according to the selected scheme.

        - 'zeros': set weights to zeros (:func:`zeros <gfdl.weights.zeros>`).

        - 'range': set weights to normalized np.arange
          (:func:`range <gfdl.weights.range>`).

        - 'uniform': uniform distribution (:func:`uniform <gfdl.weights.uniform>`).

        - 'he_uniform': He uniform distribution
          (:func:`he_uniform <gfdl.weights.he_uniform>`).

        - 'lecun_uniform': Lecun uniform distribution
          (:func:`lecun_uniform <gfdl.weights.lecun_uniform>`).

        - 'glorot_uniform': Glorot uniform distribution
          (:func:`glorot_uniform <gfdl.weights.glorot_uniform>`).

        - 'normal': Normal distribution (:func:`normal <gfdl.weights.normal>`).

        - 'he_normal': He normal distribution
          (:func:`he_normal <gfdl.weights.he_normal>`).

        - 'lecun_normal': Lecun normal distribution
          (:func:`lecun_normal <gfdl.weights.lecun_normal>`).

        - 'glorot_normal': Glorot normal distribution
          (:func:`glorot_normal <gfdl.weights.glorot_normal>`).

    direct_links : bool, default=True
        Whether to connect input layer to output nodes.

        When set to False, only the hidden-layer activations are used, corresponding
        to the Extreme Learning Machine (ELM) architecture.

    seed : int, RandomState instance, default=None
        Determines random number generation for weights and bias
        initialization.
        Pass an int for reproducible results across multiple function calls.
        See :term:`Glossary <random_state>`.

    reg_alpha : float, default=None
        Amount of ridge shrinkage to apply in order to improve
        conditioning during Ridge regression. When set to zero or `None`,
        model uses direct solve using Moore-Penrose Pseudo-Inverse.

    rtol : float, default=None
        Cutoff for small singular values for the Moore-Penrose
        pseudo-inverse. Only applies when ``reg_alpha=None``.
        When ``rtol=None``, the array API standard default for
        ``pinv`` is used.

    Attributes
    ----------
    n_features_in_ : int
        Number of features seen during :term:`fit`.

    W_ : list of ndarray of shape (n_layers,)
        Weight matrices of the hidden layers. The ith element in the list represents the
        weight matrix corresponding to layer i.

    b_ : list of ndarray of shape (n_layers,)
        Bias vectors of the hidden layers. The ith element in the list represents the
        bias term corresponding to layer i.

    coeff_ : ndarray of shape (n_features_out, n_outputs)
        Output weight matrix learned by the fit method.

    See Also
    --------
    GFDLClassifier : Classifier variant for the RVFL architecture.

    Examples
    --------
    >>> from gfdl.model import GFDLRegressor
    >>> from sklearn.datasets import make_regression
    >>> from sklearn.model_selection import train_test_split
    >>> X, y = make_regression(n_samples=200, n_features=20, random_state=1)
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y,
    ...                                                     random_state=1)
    >>> regr = GFDLRegressor(seed=1)
    >>> regr.fit(X_train, y_train)
    GFDLRegressor(seed=1)
    >>> regr.predict(X_test[:2])
    array([  18.368, -278.014])
    """
    def __init__(
        self,
        hidden_layer_sizes: np.typing.ArrayLike = (100,),
        activation: str = "identity",
        weight_scheme: str = "uniform",
        direct_links: bool = True,
        seed: int = None,
        reg_alpha: float = None,
        rtol: float | None = None,
    ):
        super().__init__(hidden_layer_sizes=hidden_layer_sizes,
                       activation=activation,
                       weight_scheme=weight_scheme,
                       direct_links=direct_links,
                       seed=seed,
                       reg_alpha=reg_alpha,
                       rtol=rtol)

    def fit(self, X, y):
        """
        Train the gradient-free neural network on the training set (X, y).

        Parameters
        ----------

        X : array-like of shape (n_samples, n_features)
          The training input samples.
        y : array-like of shape (n_samples,) or (n_samples, n_outputs)
          The target values.

        Returns
        -------
        object
          The fitted estimator.
        """
        X, Y = validate_data(self, X, y, multi_output=True)
        super().fit(X, Y)
        return self

    def predict(self, X):
        """
        Predict regression target for X.

        Parameters
        ----------

        X : array-like of shape (n_samples, n_features)
          The input samples.

        Returns
        -------
        ndarray
          The predicted values. Should have shape (n_samples,) or
          (n_samples, n_outputs).
        """
        check_is_fitted(self)
        X = validate_data(self, X, reset=False)
        return super().predict(X)



"""
Activation functions for Gradient Free Deep Learning estimators.
"""

import numpy as np
import scipy


def relu(z):
    """
    The rectified linear unit activation function.

    Parameters
    ----------
    z : array_like
        Input array.

    Returns
    -------
    numpy.ndarray
        The output array with only positive values.
    """
    return np.maximum(0, z)


def tanh(z):
    """
    The hyperbolic tangent activation function.

    Parameters
    ----------
    z : array_like
        Input array.

    Returns
    -------
    numpy.ndarray
        The output array with hyperbolic tangent values.

    See Also
    --------
    numpy.tanh : The hyperbolic tangent function.
    """
    return np.tanh(z)


def sigmoid(z):
    """
    The logistic sigmoid activation function.

    Parameters
    ----------
    z : array_like
        Input array.

    Returns
    -------
    numpy.ndarray
        The output array with the function values.

    See Also
    --------
    scipy.special.expit : The logistic sigmoid function.
    """
    return scipy.special.expit(z)


def identity(z):
    """
    The identity activation function.

    Parameters
    ----------
    z : array_like
        Input array.

    Returns
    -------
    numpy.ndarray
        The input array is returned unchanged.
    """
    return z


def softmax(z):
    """
    The softmax activation function.

    Parameters
    ----------
    z : array_like
        Input array.

    Returns
    -------
    numpy.ndarray
        The output array with the function values.

    See Also
    --------
    scipy.special.softmax : The softmax function.
    """
    return scipy.special.softmax(z, axis=-1)


def softmin(z):
    """
    The softmin activation function.

    It is the softmax function applied to negative of the input values.

    Parameters
    ----------
    z : array_like
        Input array.

    Returns
    -------
    numpy.ndarray
        The output array with the function values.

    See Also
    --------
    scipy.special.softmax : The softmax function.
    """
    return softmax(-z)


def log_sigmoid(z):
    """
    The logarithm of logistic sigmoid activation function.

    Parameters
    ----------
    z : array_like
        Input array.

    Returns
    -------
    numpy.ndarray
        The output array with the function values.

    See Also
    --------
    scipy.special.log_expit : The logistic sigmoid function.
    """
    return scipy.special.log_expit(z)


def log_softmax(z):
    """
    The log softmax activation function.

    Parameters
    ----------
    z : array_like
        Input array.

    Returns
    -------
    numpy.ndarray
        The output array with the function values.

    See Also
    --------
    scipy.special.log_softmax : The logarithm of softmax function.
    """
    return scipy.special.log_softmax(z, axis=-1)


ACTIVATIONS = {
    "relu": relu,
    "tanh": tanh,
    "sigmoid": sigmoid,
    "identity": identity,
    "linear": identity,
    "softmax": softmax,
    "softmin": softmin,
    "log_sigmoid": log_sigmoid,
    "log_softmax": log_softmax,
}


def resolve_activation(activation):
    # numpydoc ignore=GL08
    name = activation.strip().lower()
    try:
        fn = ACTIVATIONS[name]
    except KeyError as e:
        allowed = sorted(ACTIVATIONS.keys())
        raise ValueError(
            f"activation='{activation}' is not supported; choose from {allowed}"
        ) from e
    return name, fn



"""
Weight functions for Gradient Free Deep Learning estimators.
"""

import numpy as np


def zeros(d, h, **kwargs):
    """
    The weight function setting all weights to zero.

    This function is useful to test out the effect of
    the data features in isolation.

    Parameters
    ----------
    d : int
        Number of features in a sample or number of neurons in the previous
        hidden layer.

    h : int
        Number of neurons in the current hidden layer.

    Returns
    -------
    ndarray or scalar
        All zeros.

    Other Parameters
    ----------------
    **kwargs : dict
        Needed for keyword arguments and compatibility with other weight function apis
        but not relevant for this function.
    """
    return np.zeros((h, d))


def uniform(d, h, *, rng, **kwargs):
    """
    The weight function returning samples drawn from uniform distribution.

    Parameters
    ----------
    d : int
        Number of features in a sample or number of neurons in the previous
        hidden layer.

    h : int
        Number of neurons in the current hidden layer.

    rng : np.random.Generator
        A NumPy random number generator instance.

    Returns
    -------
    ndarray or scalar
        Draw samples from the uniform distribution between ``[0, 1)``.

    Other Parameters
    ----------------
    **kwargs : dict
        Other keyword arguments. Placeholder for exposing distribution
        parameters later on.
    """
    return rng.uniform(0, 1, (h, d))


def range(d, h, **kwargs):
    """
    The weight function setting weights to a normalized np.arange.

    Parameters
    ----------
    d : int
        Number of features in a sample or number of neurons in the previous
        hidden layer.

    h : int
        Number of neurons in the current hidden layer.

    Returns
    -------
    ndarray or scalar
        Set the weights to normalized np.arange over the range ``[0, d*h)``.

    Other Parameters
    ----------------
    **kwargs : dict
        Needed for keyword arguments and compatibility with other weight function apis
        but not relevant for this function.
    """
    s = np.arange(d * h)
    s = np.subtract(s, np.mean(s))
    s /= np.std(s)
    s = np.nan_to_num(s)
    return s.reshape(h, d)


def he_uniform(d, h, *, rng, **kwargs):
    """
    The weight function returning samples drawn from He uniform distribution.

    Parameters
    ----------
    d : int
        Number of features in a sample or number of neurons in the previous
        hidden layer.

    h : int
        Number of neurons in the current hidden layer.

    rng : np.random.Generator
        A NumPy random number generator instance.

    Returns
    -------
    ndarray or scalar
        Draw samples from the He uniform distribution between
        ``[sqrt(6/h), sqrt(6/h))``.

    Other Parameters
    ----------------
    **kwargs : dict
        Needed for keyword arguments and compatibility with other weight function apis
        but not relevant for this function.

    Notes
    -----
    https://faroit.com/keras-docs/2.0.0/initializers/#he_uniform
    """

    limit = np.sqrt(6 / h)
    return rng.uniform(-limit, limit, (h, d))


def lecun_uniform(d, h, *, rng, **kwargs):
    """
    The weight function returning samples drawn from Lecun uniform distribution.

    Parameters
    ----------
    d : int
        Number of features in a sample or number of neurons in the previous
        hidden layer.

    h : int
        Number of neurons in the current hidden layer.

    rng : np.random.Generator
        A NumPy random number generator instance.

    Returns
    -------
    ndarray or scalar
        Draw samples from the Lecun uniform distribution between
        ``[sqrt(3/h), sqrt(3/h))``.

    Other Parameters
    ----------------
    **kwargs : dict
        Needed for keyword arguments and compatibility with other weight function apis
        but not relevant for this function.

    Notes
    -----
    https://faroit.com/keras-docs/2.0.0/initializers/#lecun_uniform
    """

    limit = np.sqrt(3 / h)
    return rng.uniform(-limit, limit, (h, d))


def glorot_uniform(d, h, *, rng, **kwargs):
    """
    The weight function returning samples drawn from Glorot uniform distribution.

    Parameters
    ----------
    d : int
        Number of features in a sample or number of neurons in the previous
        hidden layer.

    h : int
        Number of neurons in the current hidden layer.

    rng : np.random.Generator
        A NumPy random number generator instance.

    Returns
    -------
    ndarray or scalar
        Draw samples from the Glorot uniform distribution between
        ``[-sqrt(6/(d+h)), sqrt(6/(d+h)))``.

    Other Parameters
    ----------------
    **kwargs : dict
        Needed for keyword arguments and compatibility with other weight function apis
        but not relevant for this function.

    Notes
    -----
    https://faroit.com/keras-docs/2.0.0/initializers/#glorot_uniform
    """

    fan_avg = 0.5 * (d + h)
    limit = np.sqrt(3 / fan_avg)
    return rng.uniform(-limit, limit, (h, d))


def normal(d, h, *, rng, **kwargs):
    """
    The weight function returning samples drawn from normal distribution.

    Parameters
    ----------
    d : int
        Number of features in a sample or number of neurons in the previous
        hidden layer.

    h : int
        Number of neurons in the current hidden layer.

    rng : np.random.Generator
        A NumPy random number generator instance.

    Returns
    -------
    ndarray or scalar
        Draw samples from the normal distribution with
        mean ``0`` and standard deviation ``1``.

    Other Parameters
    ----------------
    **kwargs : dict
        Other keyword arguments. Placeholder for exposing distribution
        parameters later on.
    """
    return rng.normal(0, 1, (h, d))


def he_normal(d, h, *, rng, **kwargs):
    """
    The weight function returning samples drawn from He normal distribution.

    Parameters
    ----------
    d : int
        Number of features in a sample or number of neurons in the previous
        hidden layer.

    h : int
        Number of neurons in the current hidden layer.

    rng : np.random.Generator
        A NumPy random number generator instance.

    Returns
    -------
    ndarray or scalar
        Draw samples from the He normal distribution with
        mean ``0`` and standard deviation ``sqrt(2/h)``.

    Other Parameters
    ----------------
    **kwargs : dict
        Needed for keyword arguments and compatibility with other weight function apis
        but not relevant for this function.

    Notes
    -----
    https://faroit.com/keras-docs/2.0.0/initializers/#he_normal
    """

    var = np.sqrt(2 / h)
    return rng.normal(0, var, (h, d))


def lecun_normal(d, h, *, rng, **kwargs):
    """
    The weight function returning samples drawn from Lecun normal distribution.

    Parameters
    ----------
    d : int
        Number of features in a sample or number of neurons in the previous
        hidden layer.

    h : int
        Number of neurons in the current hidden layer.

    rng : np.random.Generator
        A NumPy random number generator instance.

    Returns
    -------
    ndarray or scalar
        Draw samples from the Lecun normal distribution
        with mean ``0`` and standard deviation ``sqrt(1/h)``.

    Other Parameters
    ----------------
    **kwargs : dict
        Needed for keyword arguments and compatibility with other weight function apis
        but not relevant for this function.

    Notes
    -----
    https://www.tensorflow.org/api_docs/python/tf/keras/initializers/LecunNormal
    """

    var = 1 / np.sqrt(h)
    return rng.normal(0, var, (h, d))


def glorot_normal(d, h, *, rng, **kwargs):
    """
    The weight function returning samples drawn from Glorot normal distribution.

    Parameters
    ----------
    d : int
        Number of features in a sample or number of neurons in the previous
        hidden layer.

    h : int
        Number of neurons in the current hidden layer.

    rng : np.random.Generator
        A NumPy random number generator instance.

    Returns
    -------
    ndarray or scalar
        Draw samples from the Glorot normal distribution with
        mean ``0`` and standard deviation ``sqrt(2/(d+h))``.

    Other Parameters
    ----------------
    **kwargs : dict
        Needed for keyword arguments and compatibility with other weight function apis
        but not relevant for this function.

    Notes
    -----
    https://faroit.com/keras-docs/2.0.0/initializers/#glorot_normal
    """

    fan_avg = 0.5 * (d + h)
    var = np.sqrt(1 / fan_avg)
    return rng.normal(0, var, (h, d))


WEIGHTS = {
    "zeros": zeros,
    "uniform": uniform,
    "range": range,
    "normal": normal,
    "he_uniform": he_uniform,
    "lecun_uniform": lecun_uniform,
    "glorot_uniform": glorot_uniform,
    "he_normal": he_normal,
    "lecun_normal": lecun_normal,
    "glorot_normal": glorot_normal,
}


def resolve_weight(weight):
    # numpydoc ignore=GL08
    name = weight.strip().lower()
    try:
        w = WEIGHTS[name]
    except KeyError as e:
        allowed = sorted(WEIGHTS.keys())
        raise ValueError(
            f"weight scheme='{weight}' is not supported; choose from {allowed}"
        ) from e
    return w



import numpy as np
import pytest
from sklearn.datasets import fetch_openml, make_regression
from sklearn.metrics import r2_score
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from sklearn.utils.estimator_checks import parametrize_with_checks

from gfdl.model import GFDLRegressor


@pytest.mark.parametrize("""n_samples,
                            n_targets,
                            hidden_layer_sizes,
                            activation,
                            weight_scheme,
                            reg_alpha,
                            exp_preds_shape,
                            exp_preds_median,
                            exp_preds_min,
                            exp_preds_r2""", [
    # expected values are from the graforvfl library
    (100, 10, (100,), "relu", "glorot_normal", 10, (25, 10),
     -29.31478018, -490.57518221, 0.97537085),
    (100, 10, (100,), "tanh", "uniform", 1, (25, 10),
     -43.03897314, -504.32794352, 0.98411997),
    (100, 10, (100,), "log_softmax", "uniform", 1, (25, 10),
      -30.56871963218171, -558.1388909597706, 0.9999532782125536),
    (100, 10, (100,), "log_sigmoid", "normal", 10, (25, 10),
      -19.5976250350991, -574.1699708675857, 0.9853855947182326),
    (100, 10, (1000,), "softmin", "he_uniform", 1, (25, 10),
     -57.91870287977487, -589.6707200160679, 0.9656730623177637),
    (100, 10, (1000,), "softmax", "lecun_uniform", 10, (25, 10),
     -51.938696542946786, -513.4094105001416, 0.9589931777194366),
    (100, 100, (100,), "sigmoid", "glorot_uniform", 1, (25, 100),
     -46.92889730988215, -1585.2331437646524, 0.6496204322668526),
    (100, 100, (100,), "tanh", "he_normal", 10, (25, 100),
     -5.531248709518545, -1131.5021652659007, 0.6018381457540279),
    (100, 100, (1000,), "relu", "lecun_normal", 1, (25, 100),
     -24.857674257413233, -1241.941403822942, 0.5954067650339964),
    (100, 100, (1000,), "identity", "glorot_normal", 10, (25, 100),
     -49.66037744636776, -1418.0996396366454, 0.6387637880009253),
    (1000, 10, (100,), "log_softmax", "glorot_normal", 1, (250, 10),
     -2.157983014856103, -821.8910528092026, 0.999999671320564),
    (1000, 10, (100,), "log_sigmoid", "lecun_normal", 10, (250, 10),
     -2.25281191108881, -813.3197346939389, 0.9998208055604957),
    (1000, 10, (1000,), "softmin", "he_normal", 1, (250, 10),
     -2.932635323616438, -819.9889270165279, 0.9999535335431835),
    (1000, 10, (1000,), "softmax", "glorot_uniform", 10, (250, 10),
     -3.27895924524588, -809.0526184106433, 0.9996980844468629),
    (1000, 100, (100,), "sigmoid", "lecun_uniform", 1, (250, 100),
     40.193814730616296, -2003.2760146757932, 0.9999864051131802),
    (1000, 100, (100,), "tanh", "he_normal", 10, (250, 100),
     38.349789631939906, -1968.7361166078529, 0.9984649082549426),
    (1000, 100, (1000,), "relu", "normal", 1, (250, 100),
     47.91240910167704, -2194.259205351918, 0.8620693547752554),
    (1000, 100, (1000,), "identity", "uniform", 10, (250, 100),
     39.788475103832646, -2004.3219743138504, 0.9999999882159872)
])
def test_regression_against_grafo(n_samples, n_targets, hidden_layer_sizes,
                                  activation, weight_scheme, reg_alpha,
                                  exp_preds_shape, exp_preds_median,
                                  exp_preds_min, exp_preds_r2):
    N, d = n_samples, n_targets
    RNG = 42
    X, y = make_regression(n_samples=N,
                           n_features=d,
                           n_informative=d,
                           n_targets=n_targets,
                           noise=0.0,
                           bias=0.0,
                           random_state=RNG)

    X_train, X_test, y_train, y_test = train_test_split(X, y,
                                                        test_size=0.25,
                                                        random_state=RNG)

    # Preprocessing (use the SAME scaler for all models that need it)
    scaler = StandardScaler().fit(X_train)
    X_train_s = scaler.transform(X_train)
    X_test_s = scaler.transform(X_test)

    model = GFDLRegressor(
            hidden_layer_sizes=hidden_layer_sizes,
            activation=activation,
            weight_scheme=weight_scheme,
            direct_links=1,
            seed=RNG,
            reg_alpha=reg_alpha
        )
    model.fit(X_train_s, y_train)
    actual_preds = model.predict(X_test_s)
    actual_preds_shape = actual_preds.shape
    actual_preds_median = np.median(actual_preds)
    actual_preds_min = actual_preds.min()
    actual_preds_r2 = r2_score(y_test, actual_preds)
    np.testing.assert_allclose(actual_preds_shape, exp_preds_shape)
    np.testing.assert_allclose(actual_preds_median, exp_preds_median)
    np.testing.assert_allclose(actual_preds_min, exp_preds_min)
    np.testing.assert_allclose(actual_preds_r2, exp_preds_r2)


@parametrize_with_checks([GFDLRegressor()])
def test_sklearn_api_conformance(estimator, check):
    check(estimator)


@pytest.mark.parametrize("reg_alpha, expected", [
    (0.1, 0.78550376),
    # NOTE: for Moore-Penrose, a large singular value
    # cutoff (rtol) is required to achieve reasonable R2 with
    # the Boston Housing dataset
    (None, 0.73452466),
])
def test_regression_boston(reg_alpha, expected):
    # real-world data test with multi-layer RVFL
    boston = fetch_openml(name="boston", version=1, as_frame=False)
    X, y = boston.data, boston.target.astype(float)
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2,
                                                        random_state=42,
                                                        shuffle=True)

    scaler = StandardScaler().fit(X_train)
    X_train_s = scaler.transform(X_train)
    X_test_s = scaler.transform(X_test)

    model = GFDLRegressor(
            hidden_layer_sizes=[800] * 10,
            activation="tanh",
            weight_scheme="uniform",
            direct_links=1,
            seed=0,
            reg_alpha=reg_alpha,
            rtol=1e-3,  # has no effect for `Ridge`
        )
    model.fit(X_train_s, y_train)
    y_pred = model.predict(X_test_s)
    # RandomForestRegressor() with default params scores
    # 0.8733907 here; multi-layer GFDL with above params is a bit
    # worse, but certainly better than random chance:
    actual = r2_score(y_test, y_pred)
    np.testing.assert_allclose(actual, expected)



import platform

import numpy as np
import packaging
import scipy
import sklearn
from packaging.version import Version

from .model import GFDL

__all__ = ["GFDL"]

packaging_version = Version(packaging.__version__)
min_packaging_version = "24.0"
if packaging_version < Version(min_packaging_version):
    raise ImportError(f"{packaging_version=}, but {min_packaging_version=}")

python_version = Version(platform.python_version())
min_python_version = "3.12"
if python_version < Version(min_python_version):
    raise ImportError(f"{python_version=}, but {min_python_version=}")

numpy_version = Version(np.__version__)
min_numpy_version = "2.0.0"
if numpy_version < Version(min_numpy_version):
    raise ImportError(f"{numpy_version=}, but {min_numpy_version=}")

sklearn_version = Version(sklearn.__version__)
min_sklearn_version = "1.5.0"
if sklearn_version < Version(min_sklearn_version):
    raise ImportError(f"{sklearn_version=}, but {min_sklearn_version=}")

scipy_version = Version(scipy.__version__)
min_scipy_version = "1.13.0"
if scipy_version < Version(min_scipy_version):
    raise ImportError(f"{scipy_version=}, but {min_scipy_version=}")
